import gmpy2


def rsainvalid(n, e=65537):
    if e < 3:
        return {"detected": True, "subtest": "invalid_e"}
    if e >= n:
        return {"detected": True, "subtest": "e_too_large"}
    if gmpy2.is_prime(n):
        return {"detected": True, "subtest": "prime_n"}
    return False



import unittest
import os

import badkeys

TDPATH = f"{os.path.dirname(__file__)}/data/"


class TestEcbl(unittest.TestCase):
    def test_ecbl(self):
        with open(f"{TDPATH}ec-p256-rfc-example.key") as f:
            key = f.read()
        r = badkeys.checkpubkey(key, checks=["blocklist"])
        self.assertTrue("blocklist" in r["results"])
        with open(f"{TDPATH}ed25519-rfc-example.key") as f:
            key = f.read()
        r = badkeys.checkpubkey(key, checks=["blocklist"])
        self.assertTrue("blocklist" in r["results"])
        with open(f"{TDPATH}x448-ok.key") as f:
            key = f.read()
        r = badkeys.checkpubkey(key, checks=["blocklist"])
        self.assertFalse(r["results"])


if __name__ == "__main__":
    unittest.main()



import sys
import argparse
import signal
import re
import json

from .checks import detectandcheck, allchecks
from .checks import checkrsa, checkcrt, checksshpubkey
from .allkeys import urllookup, loadextrabl
from .scanssh import scanssh
from .scantls import scantls
from .update import update_bl

MAXINPUTSIZE = 10000

count = 0

PRECRT = "-----BEGIN CERTIFICATE-----\n"
POSTCRT = "\n-----END CERTIFICATE-----\n"


def _sighandler(_signum, _handler):
    print(f"{count} keys processed", file=sys.stderr)


def _printresults(key, where, args):
    if args.json:
        print(json.dumps(key))
        return
    kn = key["type"]
    if "bits" in key:
        kn += f"[{key['bits']}]"
    if key["type"] == "unsupported":
        print(f"Warning: Unsupported key type, {where}", file=sys.stderr)
    elif key["type"] == "unparseable":
        print(f"Warning: Unparseable input, {where}", file=sys.stderr)
    elif key["type"] == "notfound":
        print(f"Warning: No key found, {where}", file=sys.stderr)
    elif args.verbose or args.all:
        if key["results"] == {}:
            print(f"{kn} key ok, {where}")
    for check, result in key["results"].items():
        sub = ""
        if "subtest" in result:
            sub = f"/{result['subtest']}"
        print(f"{check}{sub} vulnerability, {kn}, {where}")
        if args.url and "lookup" in result:
            url = urllookup(result["blid"], result["lookup"])
            if url:
                print(url)
        if args.verbose and "debug" in result:
            print(result["debug"])
        if args.verbose and "p" in result:
            print(f"RSA p {result['p']:02x}")
        if args.verbose and "q" in result:
            print(f"RSA q {result['q']:02x}")


def runcli():
    global count
    try:
        signal.signal(signal.SIGHUP, _sighandler)
    except AttributeError:  # for OSes without SIGHUP (e.g. Windows)
        pass

    ap = argparse.ArgumentParser()
    ap.add_argument(
        "infiles", nargs="*", help="Input file (certificate, csr or public key)"
    )
    ap.add_argument(
        "-c", "--checks", help="Comma-separated list of checks (default: all)"
    )
    ap.add_argument("--list", action="store_true", help="Show list of possible checks")
    ap.add_argument(
        "-m",
        "--moduli",
        action="store_true",
        help="Input file is list of RSA hex moduli",
    )
    ap.add_argument(
        "--crt-lines", action="store_true", help="Input file is list of base64 certs"
    )
    ap.add_argument(
        "--ssh-lines", action="store_true", help="Input file is list of ssh public keys"
    )
    ap.add_argument("-a", "--all", action="store_true", help="Show all keys")
    ap.add_argument("-v", "--verbose", action="store_true", help="Verbose output")
    ap.add_argument("-j", "--json", action="store_true", help="JSON output")
    ap.add_argument(
        "-u", "--url", action="store_true", help="Show private key URL if possible"
    )
    ap.add_argument("--update-bl", action="store_true", help="Update blocklist")
    ap.add_argument(
        "--update-bl-and-urls",
        action="store_true",
        help="Update blocklist and optional URL lookup list",
    )
    ap.add_argument("--extrabl", help="comma-separated list of extra blocklist files")
    ap.add_argument(
        "-t",
        "--tls",
        action="store_true",
        help="Scan TLS (pass hostnames or IPs instead of files)",
    )
    # default ports for https, smtps, imaps, pop3s, ldaps, ftps
    # and 8443 as a common non-default https port
    ap.add_argument(
        "--tls-ports",
        default="443,465,636,990,993,995,8443",
        help="TLS ports (comma-separated)",
    )
    ap.add_argument(
        "-s",
        "--ssh",
        action="store_true",
        help="Scan SSH (pass hostnames or IPs instead of files)",
    )
    ap.add_argument("--ssh-ports", default="22", help="SSH ports (comma-separated)")
    args = ap.parse_args()

    if (
        (args.moduli and args.crt_lines)
        or (args.moduli and args.ssh_lines)
        or (args.ssh_lines and args.crt_lines)
    ):
        sys.exit("Multiple input format parameters cannot be combined.")

    if (args.moduli or args.crt_lines or args.ssh_lines) and (args.tls or args.ssh):
        sys.exit("Scan modes and input file modes cannot be combined.")

    if args.update_bl_and_urls:
        update_bl(lookup=True)
        sys.exit()
    if args.update_bl:
        update_bl()
        sys.exit()

    if args.list:
        for k, v in allchecks.items():
            print(f"{k}/{v['type']} keys: {v['desc']}")
        sys.exit()

    if not args.infiles:
        ap.print_help()
        sys.exit()

    if args.extrabl:
        extrabl = args.extrabl.split(",")
        for bl in extrabl:
            loadextrabl(bl)

    if args.checks:
        userchecks = args.checks.split(",")
        for c in userchecks:
            if c not in allchecks:
                sys.exit(f"{c} is not a valid check")
    else:
        userchecks = allchecks.keys()

    if args.tls:
        ports = [int(p) for p in args.tls_ports.split(",")]
        for host in args.infiles:
            for port in ports:
                keys = scantls(host, port, userchecks)
                for k in keys:
                    _printresults(k, f"tls:{host}:{port}", args)

    if args.ssh:
        ports = [int(p) for p in args.ssh_ports.split(",")]
        for host in args.infiles:
            for port in ports:
                keys = scanssh(host, port)
                for k in keys:
                    _printresults(k, f"ssh:{host}:{port}", args)

    if args.ssh or args.tls:
        sys.exit(1)

    for fn in args.infiles:
        if fn == "-":
            f = sys.stdin
        else:
            f = open(fn, errors="replace")
        if args.moduli:
            for line in f:
                count += 1
                if line.startswith("Modulus="):
                    line = line[8:]
                n = int(line, 16)
                r = {"type": "rsa", "bits": n.bit_length()}
                r["results"] = checkrsa(n, checks=userchecks)
                _printresults(r, f"modulus {n:02x}", args)
        elif args.crt_lines:
            lno = 0
            for line in f:
                desc = f"{fn}[{lno}]"
                ll = re.split("[,; ]", line.rstrip(), maxsplit=1)
                if len(ll) == 2:
                    desc += f" {ll[1]}"
                crt = PRECRT + ll[0] + POSTCRT
                r = checkcrt(crt, checks=userchecks)
                _printresults(r, desc, args)
                lno += 1
                count += 1
        elif args.ssh_lines:
            lno = 0
            for line in f:
                desc = f"{fn}[{lno}]"
                ll = line.rstrip().split(" ", 2)
                if len(ll) == 3:
                    desc += f" {ll[2]}"
                r = checksshpubkey(line, checks=userchecks)
                _printresults(r, desc, args)
                lno += 1
                count += 1
        else:
            fcontent = f.read(MAXINPUTSIZE)
            r = detectandcheck(fcontent, checks=userchecks)
            _printresults(r, fn, args)

        if fn != "-":
            f.close()



import hashlib
import cryptography
from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import rsa, dsa, ec, dh
from cryptography.hazmat.primitives.asymmetric import ed25519, x25519
from cryptography.hazmat.primitives.asymmetric import x448, ed448
from cryptography.hazmat.primitives import serialization

from .rsakeys import fermat
from .rsakeys import pattern
from .rsakeys import roca
from .rsakeys import rsainvalid
from .rsakeys import sharedprimes
from .rsakeys import smallfactors
from .allkeys import blocklist

# List of available checks
allchecks = {
    "fermat": {
        "type": "rsa",
        "function": fermat,
        "desc": "Fermat factorization / close prime vulnerability",
    },
    "pattern": {
        "type": "rsa",
        "function": pattern,
        "desc": "Implausible repetition pattern in modulus",
    },
    "roca": {
        "type": "rsa",
        "function": roca,
        "desc": "Return of the Coopersmith Attack (ROCA) vulnerability",
    },
    "rsainvalid": {
        "type": "rsa",
        "function": rsainvalid,
        "desc": "RSA keys with invalid values",
    },
    "sharedprimes": {
        "type": "rsa",
        "function": sharedprimes,
        "desc": "Shared prime factors (batchgcd)",
    },
    "smallfactors": {
        "type": "rsa",
        "function": smallfactors,
        "desc": "Small prime factors (<=65537, usually corrupt)",
    },
    "blocklist": {
        "type": "all",
        "function": blocklist,
        "desc": "Blocklists of compromised keys",
    },
}


def _checkkey(key, checks):
    r = {}
    if isinstance(key, rsa.RSAPublicKey):
        r["type"] = "rsa"
        r["n"] = key.public_numbers().n
        r["e"] = key.public_numbers().e
        r["bits"] = r["n"].bit_length()
        r["results"] = checkrsa(r["n"], e=r["e"], checks=checks)
    elif isinstance(key, ec.EllipticCurvePublicKey):
        r["type"] = "ec"
        r["x"] = key.public_numbers().x
        r["y"] = key.public_numbers().y
        r["results"] = checkall(r["x"], checks=checks)
    elif isinstance(key, dsa.DSAPublicKey):
        r["type"] = "dsa"
        r["y"] = key.public_numbers().y
        r["bits"] = key.key_size
        r["results"] = checkall(r["y"], checks=checks)
    elif (
        isinstance(key, ed25519.Ed25519PublicKey)
        or isinstance(key, x25519.X25519PublicKey)
        or isinstance(key, x448.X448PublicKey)
        or isinstance(key, ed448.Ed448PublicKey)
    ):
        r["type"] = "ec"
        # For Ed25519 the raw key is the x coordinate
        x_b = key.public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw,
        )
        r["x"] = int.from_bytes(x_b, byteorder="big")
        # we don't need the y coordinate
        r["results"] = checkall(r["x"], checks=checks)
    elif isinstance(key, dh.DHPublicKey):
        r["type"] = "dh"
        try:
            r["y"] = key.public_numbers().y
        except ValueError:
            # happens with e.g. very small (<512) DH keys
            return {"type": "unparseable", "results": {}}
        r["results"] = checkall(r["y"], checks=checks)
    else:
        r["type"] = "unsupported"
        r["results"] = {}
    spki = key.public_bytes(
        serialization.Encoding.DER,
        serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    r["spkisha256"] = hashlib.sha256(spki).hexdigest()
    return r


def checkrsa(n, e=65537, checks=allchecks.keys()):
    results = {}
    for check in checks:
        callcheck = allchecks[check]["function"]
        if allchecks[check]["type"] == "rsa":
            r = callcheck(n, e=e)
        elif allchecks[check]["type"] == "all":
            r = callcheck(n)
        else:
            continue
        if r is not False:
            results[check] = r
    return results


def checkall(x, checks=allchecks.keys()):
    results = {}
    for check in checks:
        if allchecks[check]["type"] != "all":
            continue
        callcheck = allchecks[check]["function"]
        r = callcheck(x)
        if r is not False:
            results[check] = r
    return results


def checkpubkey(rawkey, checks=allchecks.keys()):
    try:
        key = serialization.load_pem_public_key(rawkey.encode())
    except ValueError:
        # happens e.g. on partial inputs
        return {"type": "unparseable", "results": {}}
    return _checkkey(key, checks)


def checkprivkey(rawkey, checks=allchecks.keys()):
    try:
        priv = serialization.load_pem_private_key(rawkey.encode(), password=None)
    except ValueError:
        # happens on invalid values, e.g. p=q
        return {"type": "unparseable", "results": {}}
    except cryptography.exceptions.UnsupportedAlgorithm:
        # happens e.g. with unsupported curves
        return {"type": "unparseable", "results": {}}
    except TypeError:
        # happens on keys with passwords
        return {"type": "unparseable", "results": {}}
    return _checkkey(priv.public_key(), checks)


def checkcrt(rawcert, checks=allchecks.keys()):
    try:
        crt = x509.load_pem_x509_certificate(rawcert.encode())
    except (ValueError, cryptography.x509.base.InvalidVersion):
        return {"type": "unparseable", "results": {}}
    try:
        return _checkkey(crt.public_key(), checks)
    except cryptography.exceptions.UnsupportedAlgorithm:
        # happens e.g. with PSS keys
        return {"type": "unsupported", "results": {}}
    except (ValueError, NotImplementedError):
        # happens e.g. with ECDSA custom curves
        return {"type": "unsupported", "results": {}}


def checkcsr(rawcsr, checks=allchecks.keys()):
    csr = x509.load_pem_x509_csr(rawcsr.encode())
    return _checkkey(csr.public_key(), checks)


def checksshprivkey(sshkey, checks=allchecks.keys()):
    try:
        pkey = serialization.load_ssh_private_key(sshkey.encode(), password=None)
    except ValueError:
        # happens e.g. on password-protected keys
        return {"type": "unsupported", "results": {}}
    except cryptography.exceptions.UnsupportedAlgorithm:
        # happens e.g. on pre-standard sk-ssh-ed25519@openssh.com keys
        return {"type": "unsupported", "results": {}}
    return _checkkey(pkey.public_key(), checks)


def checksshpubkey(sshkey, checks=allchecks.keys()):
    try:
        pkey = serialization.load_ssh_public_key(sshkey.encode())
    except ValueError:
        # happens e.g. on non-standard DSA keys (!=1024 bit)
        return {"type": "unsupported", "results": {}}
    except cryptography.exceptions.UnsupportedAlgorithm:
        # happens e.g. on pre-standard sk-ssh-ed25519@openssh.com keys
        return {"type": "unsupported", "results": {}}
    return _checkkey(pkey, checks)


def detectandcheck(inkey, checks=allchecks.keys()):
    if "-----BEGIN CERTIFICATE-----" in inkey:
        return checkcrt(inkey, checks)
    elif "-----BEGIN CERTIFICATE REQUEST-----" in inkey:
        return checkcsr(inkey, checks)
    elif "-----BEGIN PUBLIC KEY-----" in inkey:
        return checkpubkey(inkey, checks)
    elif "-----BEGIN RSA PUBLIC KEY-----" in inkey:
        return checkpubkey(inkey, checks)
    elif "-----BEGIN PRIVATE KEY-----" in inkey:
        return checkprivkey(inkey, checks)
    elif "-----BEGIN RSA PRIVATE KEY-----" in inkey:
        return checkprivkey(inkey, checks)
    elif "-----BEGIN DSA PRIVATE KEY-----" in inkey:
        return checkprivkey(inkey, checks)
    elif "-----BEGIN EC PRIVATE KEY-----" in inkey:
        return checkprivkey(inkey, checks)
    elif "-----BEGIN OPENSSH PRIVATE KEY-----" in inkey:
        return checksshprivkey(inkey, checks)
    elif inkey.startswith("ssh-") or inkey.startswith("ecdsa-"):
        return checksshpubkey(inkey, checks)
    else:
        return {"type": "notfound", "results": {}}



import hashlib
import mmap
import pathlib
import json
import sys

_blmeta = False
_bldata = False
_blextra = []


def _loadblmeta():
    global _blmeta
    mlist = {}
    cachedir = str(pathlib.Path.home()) + "/.cache/badkeys/"
    try:
        with open(f"{cachedir}badkeysdata.json") as f:
            jdata = json.loads(f.read())
    except FileNotFoundError:
        sys.exit("blocklist metadata not found, you need to run --update-bl")
    for bl in jdata["blocklists"]:
        blid = int(bl["id"])
        mlist[blid] = bl
    _blmeta = mlist


def _checkbl(bldata, s256trunc):
    fbegin = 0
    fend = (len(bldata) // 16) - 1
    while fbegin <= fend:
        fmiddle = (fbegin + fend) // 2
        val = bldata[fmiddle * 16 : fmiddle * 16 + 15]
        if s256trunc == val:
            bl_id = int(bldata[fmiddle * 16 + 15])
            if bl_id in _blmeta:
                subtest = _blmeta[bl_id]["name"]
            else:
                subtest = f"id{bl_id}"
            lhash = s256trunc[0:8].hex()
            return {
                "detected": True,
                "subtest": subtest,
                "blid": bl_id,
                "lookup": lhash,
                "debug": "Truncated Hash: %s" % s256trunc.hex(),
            }
        if s256trunc > val:
            fbegin = fmiddle + 1
        else:
            fend = fmiddle - 1

    return False


def blocklist(inval):
    global _blmeta, _bldata

    if not _blmeta:
        _loadblmeta()

    if not _bldata:
        cachedir = str(pathlib.Path.home()) + "/.cache/badkeys/"
        try:
            with open(f"{cachedir}blocklist.dat", "rb") as f:
                _bldata = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
        except FileNotFoundError:
            sys.exit("blocklist.dat not found, you need to run --update-bl")

    inval_b = inval.to_bytes((inval.bit_length() + 7) // 8, byteorder="big")

    s256trunc = hashlib.sha256(inval_b).digest()[:15]

    ret = _checkbl(_bldata, s256trunc)

    if not ret and _blextra != []:
        for bl in _blextra:
            ret = _checkbl(bl, s256trunc)

    return ret


def loadextrabl(fpath):
    with open(fpath, "rb") as f:
        _blextra.append(mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ))


def urllookup(blid, lhash, type="show"):
    global _blmeta
    try:
        from binary_file_search.BinaryFileSearch import BinaryFileSearch
    except ModuleNotFoundError:
        sys.stderr.write("ERROR: URL lookup failed, needs binary_file_search module\n")
        return False

    if not _blmeta:
        _loadblmeta()

    lfile = str(pathlib.Path.home()) + "/.cache/badkeys/lookup.txt"

    try:
        with BinaryFileSearch(lfile, sep=";", string_mode=True) as bfs:
            x = bfs.search(lhash)
    except FileNotFoundError:
        sys.stderr.write("ERROR: lookup.txt not found, run --update-bl-and-urls\n")
        return False
    except KeyError:
        sys.stderr.write("ERROR: URL lookup failed, not found\n")
        return False
    d = _blmeta[blid]
    showurl = f"https://github.com/{d['repo']}/blob/{d['path']}/{x[0][1]}"
    rawurl = f"https://raw.githubusercontent.com/{d['repo']}/{d['path']}/{x[0][1]}"
    if type == "show":
        return showurl
    if type == "raw":
        return rawurl
    if type == "both":
        return [showurl, rawurl]



__all__ = ["fermat", "pattern", "roca", "rsainvalid", "sharedprimes", "smallfactors"]
from .fermat import fermat
from .pattern import pattern
from .roca import roca
from .rsainvalid import rsainvalid
from .sharedprimes import sharedprimes
from .smallfactors import smallfactors



import gmpy2

MAX_PRIME = 65537

_smallprimes = None


def smallfactors(n, e=0):
    global _smallprimes

    # Generate product of all primes <= MAX_PRIME.
    # We calculate this once per program run, we could precalculate the
    # constant, but it's fast enough to calculate on the fly.
    if _smallprimes is None:
        sp = 2
        for i in range(3, MAX_PRIME + 1, 2):
            if gmpy2.is_prime(i):
                sp *= i
        _smallprimes = sp

    factor = gmpy2.gcd(_smallprimes, n)
    if factor != 1:
        return {"p": int(factor), "q": int(n // factor)}
    return False



import gmpy2


def fermat(n, e=65535):
    tries = 100

    a = gmpy2.isqrt(n)

    c = 0
    while not gmpy2.is_square(a**2 - n):
        a += 1
        c += 1
        if c > tries:
            return False
    bsq = a**2 - n
    b = gmpy2.isqrt(bsq)
    # Technically it should not matter whether p or q is larger,
    # but some implementations seem to prefer p>q, therefore we
    # use p as the larger prime
    p = a + b
    q = a - b
    debugmsg = f"Fermat factorization after {c} rounds, b={b}"
    return {"p": int(p), "q": int(q), "a": int(a), "b": int(b), "debug": debugmsg}



import contextlib
import io

from .checks import checksshpubkey, allchecks


def scanssh(host, port=22, checks=allchecks.keys()):
    import paramiko

    allkeytypes = paramiko.Transport._preferred_keys

    keys = []
    for keytype in ["rsa", "dss", "ecdsa", "ed25519"]:
        try:
            xnot = [x for x in allkeytypes if keytype not in x]

            transp = paramiko.Transport(
                f"{host}:{port}", disabled_algorithms={"keys": xnot}
            )

            with contextlib.redirect_stderr(io.StringIO()):
                transp.connect()
            key = transp.get_remote_server_key()
            transp.close()

            pubsshkey = f"{key.get_name()} {key.get_base64()}"
            keys.append(checksshpubkey(pubsshkey, checks=checks))
        except paramiko.ssh_exception.IncompatiblePeer:
            # if we can't connect with this key type
            # we try the next
            continue
        except paramiko.ssh_exception.SSHException:
            # if we can't connect at all we don't try again
            break

    return keys



import unittest
import os

import badkeys

TDPATH = f"{os.path.dirname(__file__)}/data/"


class TestRoca(unittest.TestCase):
    def test_roca(self):
        with open(f"{TDPATH}rsa-roca.key") as f:
            key = f.read()
        r = badkeys.checkpubkey(key, checks=["roca"])
        self.assertTrue("roca" in r["results"])
        with open(f"{TDPATH}rsa-ok.key") as f:
            key = f.read()
        r = badkeys.checkpubkey(key, checks=["roca"])
        self.assertFalse(r["results"])


if __name__ == "__main__":
    unittest.main()



import unittest
import os

import badkeys

TDPATH = f"{os.path.dirname(__file__)}/data/"


class TestFermat(unittest.TestCase):
    def test_fermat(self):
        # For this test only we are testing several
        # input formats. Also the test keys differ in
        # the distance of the "close" primes to test
        # different rounds of the fermat check.

        with open(f"{TDPATH}rsa-fermat.crt") as f:
            inp = f.read()
        r = badkeys.checkcrt(inp, ["fermat"])
        self.assertTrue("fermat" in r["results"])
        r = badkeys.detectandcheck(inp, checks=["fermat"])
        self.assertTrue("fermat" in r["results"])

        with open(f"{TDPATH}rsa-fermat.csr") as f:
            inp = f.read()
        r = badkeys.checkcsr(inp, ["fermat"])
        self.assertTrue("fermat" in r["results"])
        r = badkeys.detectandcheck(inp, checks=["fermat"])
        self.assertTrue("fermat" in r["results"])

        with open(f"{TDPATH}rsa-fermat-pkcs1.key") as f:
            inp = f.read()
        r = badkeys.checkpubkey(inp, ["fermat"])
        self.assertTrue("fermat" in r["results"])
        r = badkeys.detectandcheck(inp, checks=["fermat"])
        self.assertTrue("fermat" in r["results"])

        with open(f"{TDPATH}rsa-fermat-pkcs8.key") as f:
            inp = f.read()
        r = badkeys.checkpubkey(inp, checks=["fermat"])
        self.assertTrue("fermat" in r["results"])
        r = badkeys.detectandcheck(inp, checks=["fermat"])
        self.assertTrue("fermat" in r["results"])

        with open(f"{TDPATH}rsa-fermat-hexmodulus.txt") as f:
            inp = f.read()
        n = int(inp, 16)
        r = badkeys.checkrsa(n=n, checks=["fermat"])
        self.assertTrue("fermat" in r)
        # check that Fermat factorization works
        p = r["fermat"]["p"]
        q = r["fermat"]["q"]
        self.assertEqual(p * q, n)

        with open(f"{TDPATH}rsa-ok.key") as f:
            key = f.read()
        r = badkeys.checkpubkey(key, checks=["fermat"])
        self.assertFalse(r["results"])


if __name__ == "__main__":
    unittest.main()



import re

# Find suspicious patterns of 16 repeating bytes
_prex = re.compile(r"(..)\1{15}")


def pattern(n, e=0):
    r = _prex.search(f"{n:02x}")
    if r:
        return {"detected": True}
    return False



import gmpy2
from importlib.resources import open_binary

_moduli = {}

_supported_bits = [512, 768, 1024, 2048, 4096]


def sharedprimes(n, e=0):
    global _moduli
    bits = n.bit_length()
    if bits not in _supported_bits:
        return False

    if bits not in _moduli:
        with open_binary("badkeys.keydata", f"primes{bits}.dat") as f:
            _moduli[bits] = gmpy2.from_binary(f.read())

    breakme = gmpy2.gcd(n, _moduli[bits])
    if breakme == 1:
        return False
    if gmpy2.is_prime(breakme):
        p = breakme
        q = n // p
        if n == (p * q):
            return {"detected": True, "p": int(p), "q": int(q)}
    # Factoring failed
    return {"detected": True}



import urllib.request
import os.path
import pathlib
import json
import lzma
import hashlib
import sys


def update_bl(lookup=False):
    UPDATEURL = "https://update.badkeys.info/"
    BKFORMAT = 0

    cachedir = str(pathlib.Path.home()) + "/.cache/badkeys/"
    if not os.path.exists(cachedir):
        pathlib.Path(cachedir).mkdir(parents=True)

    jurl = f"{UPDATEURL}/v{BKFORMAT}/badkeysdata.json"
    bkdata = urllib.request.urlopen(jurl).read().decode()
    bkdata_old = ""
    if os.path.exists(f"{cachedir}badkeysdata.json"):
        with open(f"{cachedir}badkeysdata.json") as f:
            bkdata_old = f.read()

    data = json.loads(bkdata)
    if data["bkformat"] != BKFORMAT:
        sys.exit("ERROR: Wrong format")

    if bkdata == bkdata_old:
        print("No new data")
    else:
        print("Writing new badkeysdata.json...")
        with open(f"{cachedir}badkeysdata.json", "w") as f:
            f.write(bkdata)

    # starting with blocklist.dat
    oldbl_sha256 = ""
    if os.path.exists(f"{cachedir}blocklist.dat"):
        with open(f"{cachedir}blocklist.dat", "rb") as f:
            oldbl_sha256 = hashlib.sha256(f.read()).hexdigest()

    if oldbl_sha256 != data["blocklist_sha256"]:
        print("Downloading blocklist.dat...")
        xzblocklist = urllib.request.urlopen(data["blocklist_url"]).read()
        blocklist = lzma.decompress(xzblocklist)
        newbl_sha256 = hashlib.sha256(blocklist).hexdigest()
        if newbl_sha256 != data["blocklist_sha256"]:
            sys.exit("ERROR: SHA256 hash of downloaded blocklist.dat does not match")
        with open(f"{cachedir}blocklist.dat", "wb") as f:
            f.write(blocklist)

    # starting with lookup.txt
    oldlu_sha256 = ""
    if os.path.exists(f"{cachedir}lookup.txt"):
        with open(f"{cachedir}lookup.txt", "rb") as f:
            oldlu_sha256 = hashlib.sha256(f.read()).hexdigest()
        if not lookup and (oldlu_sha256 != data["lookup_sha256"]):
            print("WARNING: Old lookup.txt file found.")
            print("You may want to run --update-bl-and-urls")

    if lookup and (oldlu_sha256 != data["lookup_sha256"]):
        print("Downloading lookup.txt...")
        xzlookup = urllib.request.urlopen(data["lookup_url"]).read()
        lookup = lzma.decompress(xzlookup)
        newlu_sha256 = hashlib.sha256(lookup).hexdigest()
        if newlu_sha256 != data["lookup_sha256"]:
            sys.exit("ERROR: SHA256 hash of downloaded lookup.txt does not match")
        with open(f"{cachedir}lookup.txt", "wb") as f:
            f.write(lookup)



import unittest
import os

import badkeys

TDPATH = f"{os.path.dirname(__file__)}/data/"


class TestFailures(unittest.TestCase):
    def test_brokencrt(self):
        brokencrt = "-----BEGIN CERTIFICATE-----\nMII"
        r = badkeys.detectandcheck(brokencrt)
        self.assertEqual(r["type"], "unparseable")

    def test_empty(self):
        r = badkeys.detectandcheck("")
        self.assertEqual(r["type"], "notfound")

    def test_unsupportedssh(self):
        r = badkeys.detectandcheck("ssh-invalid invalid")
        self.assertEqual(r["type"], "unsupported")


if __name__ == "__main__":
    unittest.main()



import unittest
import os

import badkeys

TDPATH = f"{os.path.dirname(__file__)}/data/"


class TestSmallprimes(unittest.TestCase):
    def test_smallprimes(self):
        with open(f"{TDPATH}rsa-e1.key") as f:
            key = f.read()
        r = badkeys.checkpubkey(key, checks=["rsainvalid"])
        self.assertTrue("rsainvalid" in r["results"])
        self.assertEqual("invalid_e", r["results"]["rsainvalid"]["subtest"])

        with open(f"{TDPATH}rsa-nprime.key") as f:
            key = f.read()
        r = badkeys.checkpubkey(key, checks=["rsainvalid"])
        self.assertTrue("rsainvalid" in r["results"])
        self.assertEqual("prime_n", r["results"]["rsainvalid"]["subtest"])

        with open(f"{TDPATH}rsa-ok.key") as f:
            key = f.read()
        r = badkeys.checkpubkey(key, checks=["rsainvalid"])
        self.assertFalse(r["results"])


if __name__ == "__main__":
    unittest.main()



__all__ = [
    "allchecks",
    "checkrsa",
    "checkpubkey",
    "checkprivkey",
    "checkcrt",
    "checkcsr",
    "checksshpubkey",
    "detectandcheck",
    "scanssh",
]
from .checks import (
    allchecks,
    checkrsa,
    checkpubkey,
    checkprivkey,
    checkcrt,
    checkcsr,
    checksshpubkey,
    detectandcheck,
)
from .scanssh import scanssh
